import time

from ky3m.report import Report
from ky3m.common.str_base import str_base
from ky3m.common.latin import latin_36_human

from . import pickler as pcl

rel_dir = '\\saved_jars'


# get saved ids
def get_ids(_rep: Report) -> dict:
    ids = pcl.recall('saved_ids', _rep)
    if ids:
        _rep.record(f'Got saved ids! length: {len(ids)}', __name__)
        return ids
    else:
        _rep.record('Did not get any saved ids!', __name__)
        return {}


# simplify ids (for human)
def get_ids_simplified(ids: list, _rep: Report) -> dict:
    if len(ids) > 1:
        # equate length
        max_len = max(len(i) for i in ids)
        ids = list(i.rjust(max_len, 'U') for i in ids)
        ids_o = ids.copy()
        ids = dict(zip(ids, ids))
        ids_c = ids.copy()
        while len(ids.keys()) == len(ids_c.keys()):
            ids = ids_c.copy()
            ids_c = dict(zip(list(key[1:] for key in ids_c.keys()), ids_c.values()))
        _rep.record(f'Ids list simplified! Cut: {len(ids_o[0]) - len(list(ids.keys())[0])}', __name__)
        return dict(zip((val.strip('U') for val in ids.values()), ids.keys()))
    elif len(ids) == 1:
        _rep.record(f'Id simplified to one char!', __name__)
        return {ids[0]: ids[0][-1]}
    else:
        _rep.record(f'Nothing to simplify!', __name__)
        return {}


# anti get_ids_simplified()
def get_true_id(ids: dict, sid, _rep: Report) -> str | None:
    ids = dict(zip(ids.values(), ids.keys()))
    try:
        tid = ids[sid]
        _rep.record(f'ID converted to a true id! simple ID: {sid} true id: {tid}', __name__)
    except KeyError:
        tid = None
        _rep.record(f'ID could not be converted to a true ID! simple id: {sid}', __name__)
    return tid


def check_id_free(x: str, _rep: Report) -> bool:
    return not bool(pcl.recall(x, _rep, rel_dir))


# save jar
def save(jar: bytes, name: str, _rep: Report) -> str:
    # get id
    saved_id = str_base(abs(hash(jar + bytes([time.time_ns() % 256]))), latin_36_human)
    while not check_id_free(saved_id, _rep):
        # when id is not free, can happen theoretically (super ultra mega rare)
        # chance for 32-bit system: ~1/2.1B, chance for 64-bit systems: ~1/9.2Q
        # also can happen if user saves the same file several times (1/256 chance)
        saved_id = str_base(abs(hash(jar + bytes([hash(saved_id) + time.time_ns()]))), latin_36_human)
    _rep.record(f'Unique id provided! ({saved_id.upper()})', __name__)

    # save jar
    pcl.remember(jar, saved_id, _rep, rel_dir)
    _rep.record(f'.jar saved! saved_id: {saved_id.upper()}', __name__)

    # save id
    ids = get_ids(_rep)
    ids[saved_id] = name
    pcl.remember(ids, 'saved_ids', _rep)

    saved_id = get_ids_simplified(list(ids.keys()), _rep)[saved_id]

    return saved_id


# load saved jar
def load(saved_id: str, _rep: Report) -> tuple[bytes, bool]:
    # load .jar
    jar = pcl.recall(saved_id, _rep, '\\saved_jars')

    if jar:
        _rep.record(f'.jar loaded! saved_id: {saved_id.upper()}', __name__)
        return jar, True
    else:
        _rep.record(f'unable to load .jar! saved_id: {saved_id.upper()}', __name__)
        return b'', False


# delete saved jar
def delete(saved_id: str, _rep: Report) -> bool:
    # unsave .jar
    status = pcl.forget(saved_id, _rep, '\\saved_jars')
    _rep.record(f'.jar unsaved! saved_id: {saved_id.upper()}', __name__)

    # unsave id
    ids = get_ids(_rep)
    try:
        ids.pop(saved_id)
        pcl.remember(ids, 'saved_ids', _rep)
        _rep.record(f'id unsaved! saved_id: {saved_id.upper()}', __name__)
    except KeyError:
        _rep.record(f'id not found! saved_id: {saved_id.upper()}', __name__)
    return status



class Report:
    def __init__(self, method_name: str) -> None:
        self.__method_name = method_name
        self.__log = []
        self.__result = ''

    def __str__(self):
        return self.__result

    def record(self, text: str, name: str) -> None:
        n = '\n'
        nt = '\n\t'
        self.__log.append(f'[{name}] {text.replace(n, nt)}')

    def describe(self, text: str) -> None:
        self.__result = text

    @property
    def result(self):
        return self.__result

    @result.setter
    def result(self, value):
        self.__result = value

    @property
    def log(self):
        return self.__log



import ky3m.methods
from ky3m import methods
from ky3m.common.latin import *


# methods.py adapter
def use_method(_input_data):
    report = None

    # "m" := "method"
    m_fields = _input_data.split(' ')
    m_name = m_fields[0]

    # specification definition
    try:
        m_spec = m_fields[1:]
    except IndexError:  # if no spec. provided
        m_spec = []

    try:
        if m_name.casefold() in dir(methods):  # check for method existence
            m = getattr(methods, m_name.casefold())
            report = m(tuple(m_spec))  # method call, report.Report class returned
            if report is None:
                print('Nothing to report!')
            else:
                print(report)
        else:
            print(f'Method {m_name} does not exist!')
    except ky3m.methods.InvalidSpecification:
        print(f'Method {m_name} accepts a different specification format!')
    except NotImplementedError:
        print(f'Method {m_name} is not implemented!')

    return report


# developer function
def use_method_adv(_input_data):
    report = use_method(_input_data)

    # check log for presence
    if report:
        log = ''
        for record in report.log:
            log += record + '\n'
        print(f'\nLog intercepted:\n{log}')
    else:
        pass


def main():
    while True:
        print('\nKy3M :> ', end='')
        input_data = input().strip()

        # for METHODS (if first word is capitalized)
        if all(char in latin_upper for char in input_data.split(' ')[0]):
            use_method(input_data)

        # for METHODS (for developers)
        elif (all(char in latin_upper for char in input_data[3:].strip().split(' ')[0]) and
              input_data.split(' ')[0][:3] == 'adv'):
            use_method_adv(input_data[3:])

        else:
            print('Unknown syntax!')


if __name__ == '__main__':
    main()



import zipfile
import io
import json

import toml

from ky3m.report import Report

from .mcmod import *


def extract_info(jar: bytes, _rep: Report) -> MCMod | None:
    with zipfile.ZipFile(io.BytesIO(jar)) as zf:
        name_list = zf.namelist()

        mcmod_info_path = ''.join(name if 'mcmod.info' in name else '' for name in name_list)
        if mcmod_info_path:  # MC Forge (1.13]
            loader = 'Forge'

            data = json.loads(zf.read(mcmod_info_path))
            _rep.record('extracted mcmod.info!', __name__)

            # format differences possible
            try:
                mod_id = data[0]['modid']
                mc_version = data[0]['mcversion']
                name = data[0]['name']
                version = data[0]['version']
                description = data[0]['description']
            except KeyError:  # can happen in some cases
                mod_id = data['modList'][0]['modid']
                mc_version = data['modList'][0]['mcversion']
                name = data['modList'][0]['name']
                version = data['modList'][0]['version']
                description = data['modList'][0]['description']

            # bad initial building mod case
            mc_version = mc_version.replace('${mcversion}', 'Unknown Version')
            version = version.replace('${version}', 'Unknown Version')

            # mod class initialization
            mod = MCMod(mod_id, mc_version)
            mod.name = name
            mod.version = version
            mod.description = description
            mod.loader = loader

            _rep.record(f'mod class provided! modid: {mod_id}, mcversion: {mc_version}', __name__)

            return mod

        elif 'META-INF/mods.toml' in name_list:  # MC Forge [1.14)
            loader = 'Forge'

            # replace() below was used because toml.py doesn't properly handle \r
            data = toml.loads(zf.read('META-INF/mods.toml').decode('UTF-8').replace('\r', ''))
            _rep.record('extracted mods.toml!', __name__)

            mod_id = data['mods'][0]['modId']
            name = data['mods'][0]['displayName']
            version = data['mods'][0]['version']
            # ${file.jarVersion} case
            # replace "${file.jarVersion}" with META-INF/MANIFEST.MF "Implementation-Version" field
            # otherwise leave "Unknown Version"
            try:
                version = version.replace('${file.jarVersion}',
                                          list(line.removeprefix('Implementation-Version: ')
                                               for line in zf.read('META-INF/MANIFEST.MF').decode('UTF-8').split('\r\n')
                                               if line.startswith('Implementation-Version'))[0])
            except IndexError:
                version = 'Unknown Version'
            description = data['mods'][0]['description']

            try:
                mc_version = list(dep['versionRange']
                                  for dep in data['dependencies'][mod_id]
                                  if dep['modId'] == 'minecraft')[0]  # str
            except KeyError:  # can happen at some cases
                mc_version = 'Unknown Version'

            # mod class initialization
            mod = MCMod(mod_id, mc_version)
            mod.name = name
            mod.version = version
            mod.description = description
            mod.loader = loader

            _rep.record(f'mod class provided! modid: {mod_id}, mcversion: {mc_version}', __name__)

            return mod

        elif 'fabric.mod.json' in name_list:  # MC Fabric
            loader = 'Fabric'

            data = json.loads(zf.read('fabric.mod.json'))
            _rep.record('extracted fabric.mod.json!', __name__)

            mod_id = data['id']
            name = data['name']
            version = data['version']
            description = data['description']

            # mcversion stored in META-INF/MANIFEST.MF
            mc_version = list(
                line.removeprefix('Fabric-Minecraft-Version:').strip('\n').strip()
                for line in zf.read('META-INF/MANIFEST.MF').decode('UTF-8').split('\n')
                if line.startswith('Fabric-Minecraft-Version: '))[0]  # str

            # mod class initialization
            mod = MCMod(mod_id, mc_version)
            mod.name = name
            mod.version = version
            mod.description = description
            mod.loader = loader

            _rep.record(f'mod class provided! modid: {mod_id}, mcversion: {mc_version}', __name__)

            return mod

        else:
            _rep.record('data was not extracted! (no mod data found)', __name__)
            return None



latin_lower = 'abcdefghijklmnopqrstuvwxyz'
latin_upper = latin_lower.upper()
latin = latin_lower + latin_upper
latin_36 = '0123456789' + latin_lower
latin_36_human = latin_36.replace('0', '').replace('o', '')



class MCDependency:
    def __init__(self, modid: str, version: str):
        self.__id = modid
        self.__ver = version

    @property
    def version(self):
        return self.__ver


class MCMod:
    def __init__(self, modid, mcversion):
        # TODO implement version ranges; also consider @version.setter self.version() functionality
        mcversion = mcversion

        self.__id = modid
        self.__deps = [MCDependency('minecraft', mcversion)]
        self.__name = self.__id
        self.__ver = 'unknown'  # self version, not to be confused with minecraft version
        self.__loader = 'unknown'
        self.__desc = 'Another Minecraft mod...'

    @property
    def modid(self):
        return self.__id

    @modid.setter
    def modid(self, val):
        self.__id = val

    @property
    def dependencies(self):
        return self.__deps

    @property
    def name(self):
        return self.__name

    @name.setter
    def name(self, val):
        self.__name = val

    @property
    def version(self):
        return self.__ver

    @version.setter
    def version(self, val):
        self.__ver = val

    @property
    def description(self):
        return self.__desc

    @description.setter
    def description(self, val):
        self.__desc = val

    @property
    def loader(self):
        return self.__loader

    @loader.setter
    def loader(self, val):
        self.__loader = val



import os

from ky3m.report import Report
from ky3m.common.str_base import str_base
from ky3m.common.latin import latin_36_human
from ky3m.common import paths

from . import pickler as pcl


# list all .jar files in \.minecraft\mods folder
def list_jars(_rep: Report) -> dict:
    jars = {}
    for file in os.listdir(paths.mods_path):
        if file.endswith('.jar'):
            jars[str_base(id(file), latin_36_human)] = file

    if len(jars) > 1:
        # simplify INSPECT mod keys
        jars_c = jars.copy()
        while len(jars.keys()) == len(jars_c.keys()):
            jars = jars_c.copy()
            jars_c = dict(zip(list(key[1:] for key in jars_c.keys()), jars_c.values()))
    elif len(jars) == 1:
        jars = {list(jars.keys())[0][-1]: list(jars.values())[0]}

    # save keys to pcl
    pcl.remember(jars, 'listed_ids', _rep)

    _rep.record(f'jars listed: {len(jars)}', __name__)
    return jars


# get certain .jar file from \.minecraft\mods folder
def get_jar(listed_id: str, _rep: Report) -> bytes | None:
    listed_jars = pcl.recall('listed_ids', _rep)
    if listed_jars:
        if listed_id in listed_jars:
            with open(paths.mods_path + '\\' + listed_jars[listed_id], 'rb') as file:
                jar = file.read()
            _rep.record(f'requested .jar file provided! listed id: {listed_id.upper()}', __name__)
            return jar
        else:
            _rep.record(f'requested .jar file not found! listed id: {listed_id.upper()}', __name__)
            return None
    else:
        _rep.record(f'requested .jar file is in invalid format! listed id: {listed_id.upper()}', __name__)
        return None


# create .jar file with certain content in \.minecraft\mods folder
def insert_jar(jar: bytes, name: str, _rep: Report) -> bool:
    name += '.jar'
    try:
        with open(paths.mods_path + '\\' + name, 'xb') as file:
            file.write(jar)
        _rep.record(f'provided .jar inserted! name: {name}', __name__)
        return True
    except FileExistsError:
        _rep.record(f'unable to insert provided .jar! (file with same name exists) name: {name}', __name__)
        return False


# delete certain .jar file from \.minecraft\mods folder
def delete_jar(listed_id: str, _rep: Report) -> bool:
    listed_jars = pcl.recall('listed_ids', _rep)
    try:
        name = listed_jars[listed_id]
        os.remove(paths.mods_path + '\\' + name)
        _rep.record(f'requested .jar removed! listed id: {listed_id}, name: {name}', __name__)
        return True
    except (FileNotFoundError, KeyError):
        _rep.record(f'unable to remove requested .jar! listed id: {listed_id}', __name__)
        return False



import os
import bz2
import base64
import pathlib
import pickle as pcl
from typing import Any

from ky3m.report import Report
from ky3m.common import paths

pcl_path = paths.local_path + '\\pz'
pathlib.Path(pcl_path).mkdir(parents=True, exist_ok=True)


# "save"
def remember(obj: object, save_id: str, _rep: Report, rel_dir='') -> int:
    path = pcl_path + rel_dir + '\\'
    pathlib.Path(path).mkdir(parents=True, exist_ok=True)
    path += save_id + '.pz'

    # compressing and encrypting
    data = bz2.compress(base64.a85encode(pcl.dumps(obj)))
    with open(path, 'wb') as save:
        save.write(
            b'PZ3M:>' + bytes(list(b ^ 7 for b in data))

        )
    with open(path, 'rb') as save:
        size = len(save.read())

    raw_size = len(pcl.dumps(obj))

    _rep.record(
        f'data remembered!\n'
        f'id: {save_id}, rel dir: {rel_dir}, obj type: {type(obj)}\n'
        f'raw size: {raw_size}, saved size: {size}, efficiency: {round((raw_size - size) / raw_size * 100, 1)}%',
        __name__
    )

    return size


# "load"
def recall(save_id: str, _rep: Report, rel_dir='\\') -> Any | None:
    path = pcl_path + rel_dir + '\\'
    pathlib.Path(path).mkdir(parents=True, exist_ok=True)
    path += save_id + '.pz'

    # decompressing and decrypting
    try:
        with open(path, 'rb') as save:

            data = bytes(list(b ^ 7 for b in save.read()[6:]))

        try:
            obj = pcl.loads(base64.a85decode(bz2.decompress(data)))
            _rep.record(f'data recalled! id: {save_id}, rel dir: {rel_dir}, obj type: {type(obj)}', __name__)
            return obj

        except (OSError, ValueError):
            _rep.record(f'data was not recalled (invalid file format)! id: {save_id}, rel dir: {rel_dir}', __name__)
            return None

    except FileNotFoundError:
        _rep.record(f'data was not recalled (no such data stored)! id: {save_id}, rel dir: {rel_dir}', __name__)
        return None


def forget(save_id: str, _rep, rel_dir='\\') -> bool:
    path = pcl_path + rel_dir + '\\' + save_id + '.pz'
    try:
        os.remove(path)
        _rep.record(f'data forgot! id: {save_id}, rel dir: {rel_dir}', __name__)
        return True
    except FileNotFoundError:
        _rep.record(f'data was not forgot (no such data stored)! id: {save_id}, rel dir: {rel_dir}', __name__)
        return False



import os
import platform
import pathlib

# platform compliance setting
match platform.system():
    case 'Windows':
        appdata_loc = os.getenv('LOCALAPPDATA')
        appdata_rom = os.getenv('APPDATA')
        local_path = appdata_loc + '\\_Ky\\3M'
        minecraft_path = appdata_rom + '\\.minecraft'
    case 'Linux':
        home = os.path.expanduser("~")
        local_path = home + '\\.ky3m'
        minecraft_path = home + '\\.minecraft'
    case _:
        raise OSError(f'Incompatible platform: {platform.system()}')  # TODO (paths) Improve cross-platform

# creating dirs
pathlib.Path(local_path).mkdir(parents=True, exist_ok=True)
pathlib.Path(minecraft_path).mkdir(parents=True, exist_ok=True)

mods_path = minecraft_path + '\\mods'



import re
import time
import uuid

from .report import Report
from .services import *


class InvalidSpecification(Exception):
    pass


# used to get named specification dictionary from tuple, using auto-exception if needed
def _sep_spec(_spec: tuple, names: tuple[str] | tuple) -> dict:
    if len(_spec) == len(names):
        return dict(zip(names, _spec))
    else:
        raise InvalidSpecification()


def inspect(spec) -> Report:
    spec = _sep_spec(spec, ())  # no specification

    rep = Report('INSPECT')
    rep.record('INSPECT started!', __name__)
    jars_dict = mm_storage.list_jars(rep)  # get jars dict

    if jars_dict:
        jars = '\n'.join(list(f'{key.upper()}: {jars_dict[key]}' for key in jars_dict))
        rep.result = jars
    else:  # if jars_dict empty
        rep.result = '.jar files not found!'

    rep.record('INSPECT ended!', __name__)

    return rep
    # raise NotImplementedError('INSPECT is not implemented!')


def peer(spec) -> Report:
    spec = _sep_spec(spec, ('listed_id',))

    listed_id = spec['listed_id'].lower()  # listed id stored and used in lowercase

    rep = Report('PEER')
    rep.record('PEER started!', __name__)
    rep.record(f'PEER spec: {", ".join(spec.values())}', __name__)

    jar = mm_storage.get_jar(listed_id, rep)  # request jar

    if jar:
        mod = extractor.extract_info(jar, rep)  # extract mod info rom jar

        if mod:
            rep.result = f"""\
Minecraft {mod.name} {mod.version}:
-- {mod.description} --
ID: {mod.modid}
Minecraft compatible version: {mod.dependencies[0].version} {mod.loader}"""

        else:
            rep.result = 'This .jar is not valid minecraft mod!'

    else:
        rep.result = f'Bad listed id or wrong file format! ({spec["listed_id"]})'

    rep.record('PEER ended!', __name__)

    return rep

    # raise NotImplementedError('PEER is not implemented!')


def expel(spec) -> Report:
    spec = _sep_spec(spec, ('listed_id',))

    listed_id = spec['listed_id'].lower()  # listed id stored and used in lowercase

    rep = Report('EXPEL')
    rep.record('EXPEL started!', __name__)
    rep.record(f'EXPEL spec: {", ".join(spec.values())}', __name__)

    status = mm_storage.delete_jar(listed_id, rep)  # delete .jar
    if status:
        rep.result = f'.jar deleted successfully! (listed id: {listed_id.upper()})'
    else:
        rep.result = 'Unable to delete .jar!'

    rep.record('EXPEL ended!', __name__)

    return rep

    # raise NotImplementedError('EXPEL is not implemented!')


def adopt(spec) -> Report:
    spec = _sep_spec(spec, ('listed_id',))

    listed_id = spec['listed_id'].lower()  # listed id stored and used in lowercase

    rep = Report('ADOPT')
    rep.record('ADOPT started!', __name__)
    rep.record(f'ADOPT spec: {", ".join(spec.values())}', __name__)

    jar = mm_storage.get_jar(listed_id, rep)  # request jar

    if jar:
        mod = extractor.extract_info(jar, rep)

        saved_id = jar_keeper.save(
            jar,
            (f'{mod.name} {mod.version}' if mod
             else input('This .jar file isn\'t Minecraft valid mod! Provide name: ')),
            rep)
        rep.result = f'Requested .jar saved! ID: {saved_id.upper()}'
    else:  # if .jar not found
        rep.result = f'Bad listed id! ({spec["listed_id"]})'

    rep.record('ADOPT ended!', __name__)

    return rep

    # raise NotImplementedError('ADOPT is not implemented!')


def adopts(spec) -> Report:
    spec = _sep_spec(spec, ())  # no specification

    rep = Report('ADOPTS')
    rep.record('ADOPTS started!', __name__)
    rep.record(f'ADOPTS spec: {", ".join(spec.values())}', __name__)

    # convert id
    ids = jar_keeper.get_ids(rep)
    ids_s = jar_keeper.get_ids_simplified(list(ids.keys()), rep)

    if ids:
        rep.result = '\n'.join(f'{ids_s[i].upper()}: {ids[i]}' for i in ids)  # generate string from saved mods dict
    else:  # if ids empty
        rep.result = 'Did not get any saved ids!'

    rep.record('ADOPTS ended!', __name__)

    return rep

    # raise NotImplementedError('ADOPTS is not implemented!')


def release(spec) -> Report:
    spec = _sep_spec(spec, ('saved_id',))

    saved_id = spec['saved_id'].lower()  # saved id stored and used in lowercase

    rep = Report('RELEASE')
    rep.record('RELEASE started!', __name__)
    rep.record(f'RELEASE spec: {", ".join(spec.values())}', __name__)

    # convert id
    ids = jar_keeper.get_ids(rep)
    ids_s = jar_keeper.get_ids_simplified(list(ids.keys()), rep)
    saved_id = jar_keeper.get_true_id(ids_s, saved_id, rep)
    if not saved_id:
        rep.result = f'Bad saved id! ({spec["saved_id"]})'
        rep.record('RELEASE ended!', __name__)
        return rep

    # main part
    jar, status = jar_keeper.load(saved_id, rep)
    if status and saved_id in ids:  # saved id check
        # generate unique id
        try:
            name = time.strftime('%Y%m%d%H%M%S', time.gmtime()) + extractor.extract_info(jar, rep).modid
        except AttributeError:  # for non-mods
            name = time.strftime('%Y%m%d%H%M%S', time.gmtime()) + re.sub('[^A-Za-z0–9]', '', ids[saved_id]).casefold()
        status = mm_storage.insert_jar(jar, name, rep)  # insert
        if status:
            rep.result = f'Requested .jar released! (filename: {name}.jar)'
        else:  # logically should not happen, but you never know
            rep.result = f'Requested .jar could not be released!'
    else:
        rep.result = f'Bad saved id! ({spec["saved_id"]})'

    rep.record('RELEASE ended!', __name__)

    return rep

    # raise NotImplementedError('RELEASE is not implemented!')


def punish(spec) -> Report:
    spec = _sep_spec(spec, ('saved_id',))

    saved_id = spec['saved_id'].lower()  # saved id stored and used in lowercase

    rep = Report('PUNISH')
    rep.record('PUNISH started!', __name__)
    rep.record(f'PUNISH spec: {", ".join(spec.values())}', __name__)

    # convert id
    ids = jar_keeper.get_ids(rep)
    ids = jar_keeper.get_ids_simplified(list(ids.keys()), rep)
    saved_id = jar_keeper.get_true_id(ids, saved_id, rep)
    if not saved_id:
        rep.result = f'Bad saved id! ({spec["saved_id"]})'
        rep.record('PUNISH ended!', __name__)
        return rep

    status = jar_keeper.delete(saved_id, rep)  # delete

    if status:
        rep.result = 'Requested .jar unsaved!'
    else:
        rep.result = f'Bad saved id! ({spec["saved_id"]})'

    rep.record('PUNISH ended!', __name__)
    return rep

    # raise NotImplementedError('PUNISH is not implemented!')


def bundle(spec):
    spec = _sep_spec(spec, ('name_or_id',))

    rep = Report('BUNDLE')
    rep.record('BUNDLE started!', __name__)
    rep.record(f'BUNDLE spec: {", ".join(spec.values())}', __name__)

    try:
        bundle_id = uuid.UUID(spec['name_or_id'])
        bundle_obj = bundler.get_bundle(bundle_id, rep)

        # load saved bundles ids
        bundle_ids_saved = bundler.get_bundle_ids(rep)

        # load saved names
        names = jar_keeper.get_ids(rep)

        # for f-string
        _lf = '\n'
        _tab = '\t'

        try:
            # for output
            binds = []
            for _bind_id in bundle_obj:
                try:
                    binds.append(f'{_tab}{_bind_id.upper()}: {names[_bind_id]}')
                except KeyError:
                    binds.append(f'{_tab}{_bind_id.upper()}: UNKNOWN')
            if binds:
                binds = '\n'.join(binds)
            else:  # if no binds outputted
                binds = '\tNothing here...'

            rep.result = f'Name: {bundle_ids_saved[bundle_id.hex]}\n' \
                         f'ID: {str(bundle_id).upper()}\n' \
                         f'Binded IDs:\n' \
                         f'{binds}'
        except TypeError:
            rep.result = f'Could not find bundle with ID specified!'
        rep.record('BUNDLE ended!', __name__)
        return rep

    except ValueError:  # create new
        bundle_id = bundler.create_bundle(spec['name_or_id'], rep)
        rep.result = f'Bundle created successfully! (id: {str(bundle_id).upper()})'
        rep.record('BUNDLE ended!', __name__)
        return rep

    # raise NotImplementedError('BUNDLE is not implemented!')


def bundles(spec):
    spec = _sep_spec(spec, ())

    rep = Report('BUNDLES')
    rep.record('BUNDLES started!', __name__)

    # load saved bundles ids
    bundle_ids_saved = bundler.get_bundle_ids(rep)
    if not bundle_ids_saved:  # if found nothing saved
        rep.result = 'Unable to list bundles!'
    else:
        rep.result = '\n'.join(f'{str(uuid.UUID(key)).upper()}: {bundle_ids_saved[key]}' for key in bundle_ids_saved)

    rep.record('BUNDLES ended!', __name__)

    return rep

    # raise NotImplementedError('BUNDLES is not implemented!')


def burst(spec):
    spec = _sep_spec(spec, ('id',))

    rep = Report('BURST')
    rep.record('BURST started!', __name__)
    rep.record(f'BURST spec: {", ".join(spec.values())}', __name__)

    try:
        spec['id'] = uuid.UUID(spec['id']).hex
    except ValueError:
        rep.result = 'Wrong ID type!'
        rep.record('BURST ended!', __name__)
        return rep

    # load saved bundles ids
    bundle_ids_saved = bundler.get_bundle_ids(rep)

    name = bundle_ids_saved.pop(spec['id'], None)
    if name:
        pickler.remember(bundle_ids_saved, 'bundle_ids', rep)
        pickler.forget(spec['id'], rep, '\\bundles')
        rep.result = f'Bundle {name} bursted!'
    else:
        rep.result = f'Unable to burst bundle! (not found)'

    rep.record('BURST ended!', __name__)
    return rep

    # raise NotImplementedError('BURST is not implemented!')


def bind(spec):
    spec = _sep_spec(spec, ('id', 'saved_id',))

    saved_id = spec['saved_id'].lower()  # saved id stored and used in lowercase

    rep = Report('BIND')
    rep.record('BIND started!', __name__)
    rep.record(f'BIND spec: {", ".join(spec.values())}', __name__)

    # convert bundle id
    try:
        spec['id'] = uuid.UUID(spec['id'])
    except ValueError:
        rep.result = 'Wrong ID type!'
        rep.record('BIND ended!', __name__)
        return rep

    # convert saved id
    ids = jar_keeper.get_ids(rep)
    ids = jar_keeper.get_ids_simplified(list(ids.keys()), rep)
    saved_id = jar_keeper.get_true_id(ids, saved_id, rep)
    if not saved_id:
        rep.result = f'Bad saved id! ({spec["saved_id"]})'
        rep.record('BIND ended!', __name__)
        return rep

    # load saved bundle data
    bundle_obj = bundler.get_bundle(spec['id'], rep)
    if bundle_obj is None:
        rep.result = f'Bad bundle id! ({str(spec["id"]).upper()})'
        rep.record('BIND ended!', __name__)
        return rep

    # add id
    bundle_obj.append(saved_id)
    bundler.save_bundle(spec['id'], bundle_obj, rep)

    # load saved bundles ids
    bundle_ids_saved = bundler.get_bundle_ids(rep)

    rep.result = f'Saved id {saved_id.upper()} binded to bundle {bundle_ids_saved[spec["id"].hex]}!'

    rep.record('BIND ended!', __name__)
    return rep

    # raise NotImplementedError('BIND is not implemented!')


def detach(spec):
    spec = _sep_spec(spec, ('id', 'bind_id'))

    rep = Report('DETACH')
    rep.record('DETACH started!', __name__)
    rep.record(f'DETACH spec: {", ".join(spec.values())}', __name__)

    # convert bundle id
    try:
        spec['id'] = uuid.UUID(spec['id'])
    except ValueError:
        rep.result = 'Wrong ID type!'
        rep.record('DETACH ended!', __name__)
        return rep

    # load saved bundles ids
    bundle_ids_saved = bundler.get_bundle_ids(rep)

    try:
        name = bundle_ids_saved[spec['id'].hex]

        # remove id
        bundle_obj = bundler.get_bundle(spec['id'], rep)
        bundle_obj.remove(spec['bind_id'].casefold())
        bundler.save_bundle(spec['id'], bundle_obj, rep)

        rep.result = f'Detached {str(spec["bind_id"]).upper()} from {name}!'

    except (KeyError, ValueError):
        rep.result = 'Unable to detach!'

    rep.record('DETACH ended!', __name__)
    return rep

    # raise NotImplementedError('DETACH is not implemented!')


def apply(spec):
    spec = _sep_spec(spec, ('id',))

    rep = Report('APPLY')
    rep.record('APPLY started!', __name__)
    rep.record(f'APPLY spec: {", ".join(spec.values())}', __name__)

    # convert bundle id
    try:
        spec['id'] = uuid.UUID(spec['id'])
    except ValueError:
        rep.result = 'Wrong ID type!'
        rep.record('APPLY ended!', __name__)
        return rep

    # load saved bundle data
    bundle_obj = bundler.get_bundle(spec['id'], rep)
    if bundle_obj is None:
        rep.result = f'Bad bundle id! ({str(spec["id"]).upper()})'
        rep.record('APPLY ended!', __name__)
        return rep

    released = []  # for output
    for bind_id in bundle_obj:  # releasing

        # main part
        jar, status = jar_keeper.load(bind_id, rep)
        if status:  # saved id check
            # generate unique id
            try:
                name = time.strftime('%Y%m%d%H%M%S', time.gmtime()) + extractor.extract_info(jar, rep).modid
            except AttributeError:  # for non-mods
                name = time.strftime('%Y%m%d%H%M%S', time.gmtime()) + bind_id.casefold()
            mm_storage.insert_jar(jar, name, rep)  # insert
            released.append(f'{bind_id.upper()} released! ({name}.jar)')
        else:
            released.append(f'{bind_id.upper()} could not be released! (binded .jar file not found)')
    released = '\n'.join(released)

    # load saved bundles ids
    bundle_ids_saved = bundler.get_bundle_ids(rep)

    # get bundle name
    name = bundle_ids_saved[spec['id'].hex]

    rep.result = f'{released}\n{name} applied successfully!'
    rep.record('APPLY ended!', __name__)
    return rep

    # raise NotImplementedError('APPLY is not implemented!')



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['ky3m', 'ky3m.common', 'ky3m.services']

package_data = \
{'': ['*']}

install_requires = \
['toml>=0.10,<0.11']

setup_kwargs = {
    'name': 'ky3m',
    'version': '0.2.2',
    'description': 'Minecraft Mod Manager by Leon "Procybit" Shepelev',
    'long_description': '[](https://github.com/Procybit/Ky3M)\n\n[![GitHub Repo stars](https://img.shields.io/github/stars/Procybit/Ky3M?style=social)](https://github.com/Procybit/Ky3M)\n\n[![PyPI - Downloads](https://img.shields.io/pypi/dm/ky3m?style=for-the-badge)](https://pypi.org/project/ky3m/)\n\n[![GitHub](https://img.shields.io/github/license/Procybit/Ky3M?style=for-the-badge)](https://github.com/Procybit/Ky3M/blob/master/LICENSE)\n\n[![GitHub contributors](https://img.shields.io/github/contributors/Procybit/Ky3M?style=for-the-badge)](https://github.com/Procybit/Ky3M/blob/master/CONTRIBUTORS.md) [![GitHub commit activity](https://img.shields.io/github/commit-activity/w/Procybit/Ky3M?style=for-the-badge)](https://github.com/Procybit/Ky3M/commits/) [![GitHub issues](https://img.shields.io/github/issues-raw/Procybit/Ky3M?style=for-the-badge)](https://github.com/Procybit/Ky3M/issues)\n\n# Ky3M\n\nMinecraft Mod Manager by Leon "Procybit" Shepelev\n\n## Installing\n\nVia pip:\n\n```\npip install ky3m\n```\n\nYou can also update to the latest version:\n\n```\npip install ky3m --upgrade\n```\n\n**Make sure you have Python 3.10 or above installed!**\n\n## Program launch\n\nVia terminal:\n\n```\npython -m ky3m\n```\n\n## Using the CLI\n\nIf the program is run correctly, this should be in the terminal:\n```\nKy3M :>\n```\nAt the moment, the only thing that can be entered into the CLI is methods.\n\nMethods are simple commands that work directly with Ky3M internals.\n\nThe name of the method is written in CAPITAL LETTERS, the specification fields of the method are written with spaces:\n\n```\nKy3M :> METHOD Field_1 Field_2 ANoTheR--fiEld00;[-+^ 4thFIELD\nSomething happened...\n```\n\nIf the method name starts with the "adv" prefix, then the associated log will be displayed:\n\n```\nKy3M :> advMETHOD Field_1 Field_2 ANoTheR--fiEld00;[-+^ 4thFIELD\nSomething happened...\n\nLog intercepted!\n[ky3m.methods] METHOD started!\n[ky3m.something] Something happened! (something: True)\n[ky3m.methods] METHOD ended!\n```\n\n## Methods\n\n### INSPECT\n\nOutputs all .jar files names from Minecraft mods folder.\n\nASSIGNS A UNIQUE ID TO EACH NEWLY OUTPUT FILE.\n\n### PEER `id`\n\nOutputs information about certain .jar file in Minecraft mods folder.\n\nUses `id`  assigned by *INSPECT*.\n\n### EXPEL `id`\n\nPermanently deletes certain .jar file from Minecraft mods folder.\n\nUses `id` assigned by *INSPECT*.\n\n### ADOPT `saved_id`\n\nCopies certain .jar file from Minecraft mods folder to local library.\n\nUses `saved_id` assigned by *INSPECT*.\n\nCAN INTERRUPT CLI AND REQUEST NAME OF SAVED FILE IF NEEDED.\n\nDOESN\'T DELETE CERTAIN FILE.\n\nASSIGNS A UNIQUE SAVED ID TO EACH SAVED FILE.\n\n### ADOPTS\n\nOutputs all saved .jar files names from local library.\n\n### RELEASE `saved_id`\n\nCopies certain .jar file from local library to Minecraft mods folder.\n\nUses `saved_id` assigned by *ADOPT*.\n\n### PUNISH `saved_id`\n\nPermanently deletes certain .jar file from local library.\n\nUses `saved_id` assigned by *ADOPT*.\n\n### BUNDLE `name` | `bundle_id`\n\n*Note that | separates an alternates.*\n\nIf `name` not found, creates new bundle (Ky3M modpack) and outputs created bundle\'s ID.\n\nUses `name` that is any string.\n\nASSIGNS A UNIQUE ID TO EACH CREATED BUNDLE.\n\n**ALTERNATE**\n\nIf found bundle with specified `bundle_id`, outputs the bundle info.\n\nUses `bundle_id` that is valid UUID (in any form).\n\n### BUNDLES\n\nOutputs all created bundles\' IDs and names.\n\n### BURST `bundle_id`\n\nPermanently deletes certain bundle.\n\nUses `bundle_id` assigned by *BUNDLE*.\n\n### BIND `bundle_id` `saved_id`\n\nBinds certain .jar to a certain bundle.\n\nUses `bundle_id` assigned by *BUNDLE*.\n\nUses `saved_id` assigned by *ADOPT*.\n\nASSIGNS A BIND ID (BUNDLE LOCAL) BASED ON SAVED ID.\n\n### DETACH `bundle_id` `bind_id`\n\nDetached certain .jar from a certain bundle.\n\nUses `bundle_id` assigned by *BUNDLE*.\n\nUses `bind_id` assigned by *BIND*.\n\n### APPLY `bundle_id`\n\nReleases all binded to certain bundle .jar files (see *RELEASE*).\n\nUses `bundle_id` assigned by *BUNDLE*.\n\nDOESN\'T DELETE ANY FILES.\n\n## License\nThis project follows MIT license (see [LICENSE](LICENSE)).\n',
    'author': 'Leon "Procybit" Shepelev',
    'author_email': 'kyleusnaff@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/Procybit/Ky3M',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



import uuid

from ky3m.report import Report

from . import pickler as pcl


# load saved bundles ids
def get_bundle_ids(_rep: Report) -> dict:
    bundle_ids_saved = pcl.recall('bundle_ids', _rep)
    if not bundle_ids_saved:  # if found nothing saved
        bundle_ids_saved = {}
        _rep.record(f'bundle ids not found!', __name__)
    else:
        _rep.record(f'got bundle ids!', __name__)
    return bundle_ids_saved


# create bundle
def create_bundle(name: str, _rep: Report) -> uuid.UUID:
    # save initial bundle data
    bundle_id = uuid.uuid4().hex
    bundle_obj = []  # BINDed mods ids will be here
    pcl.remember(bundle_obj, bundle_id, _rep, '\\bundles')

    # save bundle id
    bundle_ids_saved = get_bundle_ids(_rep)
    bundle_ids_saved[bundle_id] = name
    pcl.remember(bundle_ids_saved, 'bundle_ids', _rep)

    _rep.record(f'bundle created! (name: {name}, id: {bundle_id})', __name__)

    return uuid.UUID(bundle_id)


# get bundle
def get_bundle(bundle_id: uuid.UUID, _rep: Report) -> list | None:
    bundle_obj = pcl.recall(bundle_id.hex, _rep, '\\bundles')
    if bundle_obj:
        _rep.record(f'got bundle! ({bundle_id.hex})', __name__)
    else:
        _rep.record(f'could not get bundle! ({bundle_id.hex})', __name__)
    return bundle_obj


# get bundle
def save_bundle(bundle_id: uuid.UUID, _bundle_obj: list, _rep) -> bool:
    bundle_obj = get_bundle(bundle_id, _rep)
    if bundle_obj is not None:
        pcl.remember(_bundle_obj, bundle_id.hex, _rep, '\\bundles')
        _rep.record(f'saved bundle! ({bundle_id.hex})', __name__)
        return True
    else:
        _rep.record(f'could not saved bundle! ({bundle_id.hex})', __name__)
        return False
